package org.tupeloproject.pchal2;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.tupeloproject.kernel.Context;
import org.tupeloproject.kernel.OperatorException;
import org.tupeloproject.kernel.Transformer;
import org.tupeloproject.kernel.TripleMatcher;
import org.tupeloproject.kernel.Unifier;
import org.tupeloproject.kernel.impl.MemoryContext;
import org.tupeloproject.rdf.Resource;
import org.tupeloproject.rdf.Triple;
import org.tupeloproject.rdf.Vocabulary.Rdf;
import org.tupeloproject.rdf.Vocabulary.Rdfs;
import org.tupeloproject.rdf.ntriples.NTriplesInputStream;
import org.tupeloproject.rdf.query.TransitiveClosure;
import org.tupeloproject.util.Tuple;

import org.junit.Test;

/**
 * Strategy
 */
public abstract class Strategy {
    final static String graphDir = System.getProperty("org.tupeloproject.pchal2.graphDir");
    final static String[] colorMap = new String[] {
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/-1982032298resliced2.img", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/-279574900atlas.img", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/-353907279resliced1.hdr", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/-419479583resliced4.img", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/-534739803atlas.hdr", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/1346030813resliced4.hdr", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/1690416097resliced3.img", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/480349012resliced2.hdr", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/798751608resliced3.hdr", "white", // merge
            "http://penrose.pnl.gov:8080/alfresco/webdav/workflow-2108762597/900446524resliced1.img", "white", // merge
            "http://provenance.mindswap.org/", "pink", // MINDSWAP
            "http://www.ipaw.info/challenge/", "pink", // MINDSWAP
            "http://penrose.pnl.gov:8080/alfresco/webdav/", "khaki", // SDG
            "tag:tupeloproject.org,2006:/2.0/pc2/md5sum/", "yellow1", // ES3
            "urn:uuid:", "yellow1", // ES3
            "urn:bogusFile:", "palegreen", // VisTrails
            "urn:moduleId:", "palegreen", // VisTrails
            "https://skurut1.cesnet.cz:9000/", "steelblue1", //CESNET
            "urn:challenge:", "steelblue1", // CESNET
            "urn:leadproject-org:data:", "mediumorchid3", // Karma
            "tag:tupeloproject.org,2006:/2.0/pc2/karma/", "mediumorchid3", // Karma
            "urn:lsid:net.sf.taverna:dataItem:", "grey", // MyGrid
            "urn:www.mygrid.org.uk/", "grey", // MyGrid
    };

    public abstract Set<Triple> getTriples(String resourcePath) throws Exception;

    protected String[] getFiles(String prefix, String suffix, String... infixes) {
        LinkedList<String> result = new LinkedList<String>();
        for(String infix : infixes) {
            result.add(prefix+infix+suffix);
        }
        return (String[]) result.toArray(new String[] { });
    }
    
    protected void xform(Context c, Transformer t) throws OperatorException {
        c.perform(t);
        for(Triple triple : t.getResult()) {
            //System.out.println(triple); // FIXME debug
        }
        c.addTriples(t.getResult());
    }

    static String getColor(Resource r) {
        String s = r.getString();
        for(int i = 0; i < colorMap.length; i+= 2) {
            if(s.startsWith(colorMap[i])) {
                return colorMap[i+1];
            }
        }
        return "white";
    }

    public Set<Triple> getAllTriples(String... resourcePaths) throws Exception {
        //PrintWriter pw = new PrintWriter(new FileOutputStream("/Users/futrelle/Desktop/sod.txt",true));
        Set<Triple> result = new HashSet<Triple>();
        for(String resourcePath : resourcePaths) {
            MemoryContext c = new MemoryContext();
            Set<Triple> triples = getTriples(resourcePath);
            c.addTriples(triples);
            int nTriples = c.getTriples().size();
            // find the steps and datasets
            int nSteps = 0;
            for(Resource type : new Resource[] { Ontology.STEP_TYPE, Ontology.DATASET_TYPE }) {
                Unifier u = new Unifier();
                u.setColumnNames("thing","label");
                u.addPattern("thing", Rdf.TYPE, type);
                u.addPattern("thing", Rdfs.LABEL, "label");
                c.perform(u);
                for(Tuple<Resource> row : u.getResult()) {
                    //pw.printf("%s %s \"%s\"\n",type,row.get(0),row.get(1));
                    nSteps++;
                }
            }
            System.out.printf("%d triples and %d step/dataset(s) in %s\n",nTriples,nSteps,resourcePath);
            result.addAll(triples);
        }
        //pw.close();
        return result;
    }

    static String prettifyLabel(Resource label) {
        return label.getString().replaceAll(".*[^a-zA-Z0-9_. ]","");
    }

    public void causalityGraph(Set<Triple> triples) throws OperatorException, IOException {
        MemoryContext c = new MemoryContext();
        c.addTriples(triples);
        // dump the triples to an n-triples file
        String filePrefix = getClass().getName();
        File temp = File.createTempFile(filePrefix,".nt");
        PrintWriter pw = new PrintWriter(new FileWriter(temp));
        for(Triple t : c) {
            pw.println(t);
        }
        pw.flush();
        pw.close();
        // now compute causality = step dependencies and io
        Unifier causality = new Unifier();
        causality.setColumnNames("ant","antLabel","cons","consLabel","datum","datumLabel");
        causality.addPattern("ant", Ontology.STEP_HAS_OUTPUT, "datum");
        causality.addPattern("cons", Ontology.STEP_HAS_INPUT, "datum");
        causality.addPattern("ant", Rdfs.LABEL, "antLabel");
        causality.addPattern("cons", Rdfs.LABEL, "consLabel");
        causality.addPattern("datum", Rdfs.LABEL, "datumLabel");
        c.perform(causality);
        // show computed relationships as a dot graph
        temp = File.createTempFile(filePrefix,".dot");
        pw = new PrintWriter(new FileWriter(temp));
        pw.println("digraph foo {");
        Set<String> lines = new HashSet<String>();
        for(Tuple<Resource> row : causality.getResult()) {
            int i = 0;
            Resource ant = row.get(i++);
            Resource antLabel = row.get(i++);
            Resource cons = row.get(i++);
            Resource consLabel = row.get(i++);
            Resource datum = row.get(i++);
            Resource datumLabel = row.get(i++);
            // shorten the labels if possible
            lines.add(String.format("\"%s\" [shape=box,label=\"%s\"];",ant,prettifyLabel(antLabel)));
            lines.add(String.format("\"%s\" [shape=box,label=\"%s\"];",cons,prettifyLabel(consLabel)));
            lines.add(String.format("\"%s\" [label=\"%s\"];",datum,prettifyLabel(datumLabel)));
            lines.add(String.format("\"%s\" -> \"%s\" [style=dotted,label=\"c\"];",ant,cons));
            lines.add(String.format("\"%s\" -> \"%s\" [label=\"o\"];",ant,datum));
            lines.add(String.format("\"%s\" -> \"%s\" [label=\"i\"];",datum,cons));
        }
        for(String line : lines) {
            pw.println(line);
        }
        pw.println("}");
        pw.flush();
        pw.close();
        String pathToDot = System.getProperty("org.tupeloproject.pchal2.pathToDot");
        if(pathToDot != null) {
            // do something
            String command = String.format("%s %s -Tjpg -o%s/%s.jpg",pathToDot,temp.getAbsolutePath(),graphDir,filePrefix);
            //System.out.println(command);
            Runtime.getRuntime().exec(command);
        }
    }

    public void graph(Set<Triple> triples) throws OperatorException, IOException {
        graph(getClass().getName(),triples);
    }

    public static void graph(String filePrefix, Set<Triple> triples) throws OperatorException, IOException {
        MemoryContext c = new MemoryContext();
        c.addTriples(triples);
        // dump the triples to an n-triples file
        File temp = File.createTempFile(filePrefix,".nt");
        PrintWriter pw = new PrintWriter(new FileWriter(temp));
        for(Triple t : c) {
            pw.println(t);
        }
        pw.flush();
        pw.close();
        // now compute output
        Unifier output = new Unifier();
        output.setColumnNames("step","stepLabel","datum","datumLabel");
        output.addPattern("step", Ontology.STEP_HAS_OUTPUT, "datum");
        output.addPattern("step", Rdfs.LABEL, "stepLabel");
        output.addPattern("datum", Rdfs.LABEL, "datumLabel");
        c.perform(output);
        // now compute input
        Unifier input = new Unifier();
        input.setColumnNames("step","stepLabel","datum","datumLabel");
        input.addPattern("step", Ontology.STEP_HAS_INPUT, "datum");
        input.addPattern("step", Rdfs.LABEL, "stepLabel");
        input.addPattern("datum", Rdfs.LABEL, "datumLabel");
        c.perform(input);
        // show computed relationships as a dot graph
        temp = File.createTempFile(filePrefix,".dot");
        pw = new PrintWriter(new FileWriter(temp));
        pw.println("digraph foo {");
        Set<String> lines = new HashSet<String>();
        Unifier ios[] = new Unifier[] { input, output };
        for(Unifier io : ios) {
            for(Tuple<Resource> row : io.getResult()) {
                int i = 0;
                Resource step = row.get(i++);
                String stepLabel = prettifyLabel(row.get(i++)) + "("+prettifyLabel(step)+")";
                Resource datum = row.get(i++);
                String datumLabel = prettifyLabel(row.get(i++)) + "("+prettifyLabel(datum)+")";
                // determine the colors of the boxes
                String stepColor = getColor(step);
                String datumColor = getColor(datum);
                // shorten the labels if possible
                lines.add(String.format("\"%s\" [shape=box,label=\"%s\",style=filled,color=%s];",step,stepLabel,stepColor));
                lines.add(String.format("\"%s\" [label=\"%s\",style=filled,color=%s];",datum,datumLabel,datumColor));
                //System.out.println("step "+stepLabel+"="+step);
                //System.out.println("datum "+datumLabel+"="+datum);
                if(io==input) {
                    lines.add(String.format("\"%s\" -> \"%s\" [label=\"i\"];",datum,step));
                } else {
                    lines.add(String.format("\"%s\" -> \"%s\" [label=\"o\"];",step,datum));
                }
            }
        }
        for(String line : lines) {
            pw.println(line);
        }
        pw.println("}");
        pw.flush();
        pw.close();
        String pathToDot = System.getProperty("org.tupeloproject.pchal2.pathToDot");
        if(pathToDot != null) {
            // do something
            String command = String.format("%s %s -Tjpg -o%s/%s.jpg",pathToDot,temp.getAbsolutePath(),graphDir,filePrefix);
            //System.out.println(command);
            Runtime.getRuntime().exec(command);
        }
    }

    public Set<Triple> infer(Set<Triple> groundTruth) throws OperatorException {
        return groundTruth;
    }

    public void graph(String... paths) throws Exception, IOException {
        graph(infer(getAllTriples(paths)));
    }

    public static Transformer xf(Context c, Transformer t) throws OperatorException {
        c.perform(t);
        c.addTriples(t.getResult());
        return new Transformer();
    }

    public static Context inferDependencies(Set<Triple> triples) throws Exception {
        // set up a context to hold the expanded graph
        Context c = new MemoryContext(triples);
        // infer dependencies as follows:
        // 1. if a dataset is the output of a step, it depends on the step
        Transformer t = new Transformer();
        t.addInPattern("step",Ontology.STEP_HAS_OUTPUT,"dataset");
        t.addOutPattern("dataset",Ontology.DEPENDS_ON,"step");
        t = xf(c,t);
        // 1. if a dataset is the input of a step, the step depends on it
        t.addInPattern("step",Ontology.STEP_HAS_INPUT,"dataset");
        t.addOutPattern("step",Ontology.DEPENDS_ON,"dataset");
        xf(c,t);
        return c;
    }

    protected abstract String[] getFiles(int part);

    protected String[] getFiles() {
        List<String> files = new LinkedList<String>();
        for(int i = 1; i <= 3; i++) {
            files.addAll(Arrays.asList(getFiles(i)));
        }
        return (String[]) files.toArray(new String[] { });
    }

    protected abstract Resource getAtlasXGraphic();

    @Test
    public void testGraph() throws Exception {
        graph(getFiles());
    }

    static Set<Triple> steps;
    public static Set<Triple> getSteps() {
        if(steps == null) {
            NTriplesInputStream ntis = new NTriplesInputStream
                    (Strategy.class.getResourceAsStream("/nt/steps.nt"));
            Set<Triple> s = new HashSet<Triple>();
            for(Triple t : ntis) {
                s.add(t);
            }
            System.out.printf("read %d sameAs assertions from steps.nt\n", s.size());
            steps = s;
        }
        return steps;
    }

    public static SortedSet<String> runQueryOne(Set<Triple> s, Resource seed) throws Exception {
        // now import the step equivalences and canonicalize owl:sameAs closures
        for(Triple t : getSteps()) {
            s.add(t);
        }
        Context c = new MemoryContext(s);
        c = new MemoryContext(Owlet.sameAs(c));
        // now infer the dependency graph
        c = inferDependencies(c.getTriples());
        //
        TripleMatcher tm = new TripleMatcher();
        tm.setPredicate(Ontology.DEPENDS_ON);
        c.perform(tm);
        // now compute transitive closure of dependency graph
        TransitiveClosure tc = new TransitiveClosure();
        tc.setSeed(seed);
        tc.setPredicate(Ontology.DEPENDS_ON);
        SortedSet<String> stepNames = new TreeSet<String>();
        for(Resource r : tc.close(c)) {
            Unifier u = new Unifier();
            u.setColumnNames("stepName");
            u.addPattern(r,Rdf.TYPE,Ontology.STEP_TYPE);
            u.addPattern(r,Ontology.STEP_NAME,"stepName");
            c.perform(u);
            for(Tuple<Resource> row : u.getResult()) {
                Resource label = row.get(0);
                stepNames.add(label.getString());
            }
        }
        return stepNames;
    }


    // execute query #1
    @Test
    public void testQueryOne() throws Exception {
        Set<Triple> model = infer(getAllTriples(getFiles()));
        SortedSet<String> stepNames = runQueryOne(model, getAtlasXGraphic());
        for(String stepName : stepNames) {
            System.out.println("step "+stepName);
        }
        Iterator<String> it = stepNames.iterator();
        assert it.next().equals(" 1. align_warp");
        assert it.next().equals(" 2. align_warp");
        assert it.next().equals(" 3. align_warp");
        assert it.next().equals(" 4. align_warp");
        assert it.next().equals(" 5. reslice");
        assert it.next().equals(" 6. reslice");
        assert it.next().equals(" 7. reslice");
        assert it.next().equals(" 8. reslice");
        assert it.next().equals(" 9. softmean");
        assert it.next().equals("10. slicer");
        assert it.next().equals("13. convert");
        assert stepNames.size()==11;
    }
}
